@@ -86,6 +86,20 @@ static bool hasAllowedPrefix(const std::string & s, const std::vector<std::strin
     return std::any_of(allowedPrefixes.begin(), allowedPrefixes.end(), [&](const std::string & i) { return !s.compare(0, i.size(), i); });
 }
 
+static std::string trim(std::string s)
+{
+    s.erase(s.begin(), std::find_if(s.begin(), s.end(), [](unsigned char ch) { return !std::isspace(ch); }));
+    s.erase(std::find_if(s.rbegin(), s.rend(), [](unsigned char ch) { return !std::isspace(ch); }).base(), s.end());
+
+    return s;
+}
+
+static std::string downcase(std::string s)
+{
+    std::transform(s.begin(), s.end(), s.begin(), [](unsigned char c){ return std::tolower(c); });
+    return s;
+}
+
 /* !!! G++ creates broken code if this function is inlined, don't know
    why... */
 template<ElfFileParams>
@@ -418,44 +432,83 @@ static uint64_t roundUp(uint64_t n, uint64_t m)
 
 
 template<ElfFileParams>
-void ElfFile<ElfFileParamNames>::shiftFile(unsigned int extraPages, Elf_Addr startPage)
+void ElfFile<ElfFileParamNames>::shiftFile(unsigned int extraPages, size_t startOffset)
 {
-    /* Move the entire contents of the file 'extraPages' pages
-       further. */
+    assert(startOffset >= sizeof(Elf_Ehdr));
+
     unsigned int oldSize = fileContents->size();
+    assert(oldSize > startOffset);
+
+    /* Move the entire contents of the file after 'startOffset' by 'extraPages' pages further. */
     unsigned int shift = extraPages * getPageSize();
     fileContents->resize(oldSize + shift, 0);
-    memmove(fileContents->data() + shift, fileContents->data(), oldSize);
-    memset(fileContents->data() + sizeof(Elf_Ehdr), 0, shift - sizeof(Elf_Ehdr));
+    memmove(fileContents->data() + startOffset + shift, fileContents->data() + startOffset, oldSize - startOffset);
+    memset(fileContents->data() + startOffset, 0, shift);
 
     /* Adjust the ELF header. */
     wri(hdr()->e_phoff, sizeof(Elf_Ehdr));
-    wri(hdr()->e_shoff, rdi(hdr()->e_shoff) + shift);
+    if (rdi(hdr()->e_shoff) >= startOffset)
+        wri(hdr()->e_shoff, rdi(hdr()->e_shoff) + shift);
 
     /* Update the offsets in the section headers. */
-    for (int i = 1; i < rdi(hdr()->e_shnum); ++i)
-        wri(shdrs.at(i).sh_offset, rdi(shdrs.at(i).sh_offset) + shift);
+    for (int i = 1; i < rdi(hdr()->e_shnum); ++i) {
+        size_t sh_offset = rdi(shdrs.at(i).sh_offset);
+        if (sh_offset >= startOffset)
+            wri(shdrs.at(i).sh_offset, sh_offset + shift);
+    }
+
+    int splitIndex = -1;
+    size_t splitShift = 0;
 
     /* Update the offsets in the program headers. */
     for (int i = 0; i < rdi(hdr()->e_phnum); ++i) {
-        wri(phdrs.at(i).p_offset, rdi(phdrs.at(i).p_offset) + shift);
-        if (rdi(phdrs.at(i).p_align) != 0 &&
-            (rdi(phdrs.at(i).p_vaddr) - rdi(phdrs.at(i).p_offset)) % rdi(phdrs.at(i).p_align) != 0) {
-            debug("changing alignment of program header %d from %d to %d\n", i,
-                rdi(phdrs.at(i).p_align), getPageSize());
-            wri(phdrs.at(i).p_align, getPageSize());
+        Elf_Off p_start = rdi(phdrs.at(i).p_offset);
+
+        if (p_start <= startOffset && p_start + rdi(phdrs.at(i).p_filesz) > startOffset && rdi(phdrs.at(i).p_type) == PT_LOAD) {
+            assert(splitIndex == -1);
+
+            splitIndex = i;
+            splitShift = startOffset - p_start;
+
+            /* This is the load segment we're currently extending within, so we split it. */
+            wri(phdrs.at(i).p_offset, startOffset);
+            wri(phdrs.at(i).p_memsz, rdi(phdrs.at(i).p_memsz) - splitShift);
+            wri(phdrs.at(i).p_filesz, rdi(phdrs.at(i).p_filesz) - splitShift);
+            wri(phdrs.at(i).p_paddr, rdi(phdrs.at(i).p_paddr) + splitShift);
+            wri(phdrs.at(i).p_vaddr, rdi(phdrs.at(i).p_vaddr) + splitShift);
+
+            p_start = startOffset;
+        }
+
+        if (p_start >= startOffset) {
+            wri(phdrs.at(i).p_offset, p_start + shift);
+            if (rdi(phdrs.at(i).p_align) != 0 &&
+                (rdi(phdrs.at(i).p_vaddr) - rdi(phdrs.at(i).p_offset)) % rdi(phdrs.at(i).p_align) != 0) {
+                debug("changing alignment of program header %d from %d to %d\n", i,
+                    rdi(phdrs.at(i).p_align), getPageSize());
+                wri(phdrs.at(i).p_align, getPageSize());
+            }
+        } else {
+            /* If we're not physically shifting, shift back virtual memory. */
+            if (rdi(phdrs.at(i).p_paddr) >= shift)
+                wri(phdrs.at(i).p_paddr, rdi(phdrs.at(i).p_paddr) - shift);
+            if (rdi(phdrs.at(i).p_vaddr) >= shift)
+                wri(phdrs.at(i).p_vaddr, rdi(phdrs.at(i).p_vaddr) - shift);
         }
     }
 
+    assert(splitIndex != -1);
+
     /* Add a segment that maps the new program/section headers and
        PT_INTERP segment into memory.  Otherwise glibc will choke. */
     phdrs.resize(rdi(hdr()->e_phnum) + 1);
     wri(hdr()->e_phnum, rdi(hdr()->e_phnum) + 1);
     Elf_Phdr & phdr = phdrs.at(rdi(hdr()->e_phnum) - 1);
     wri(phdr.p_type, PT_LOAD);
-    wri(phdr.p_offset, 0);
-    wri(phdr.p_vaddr, wri(phdr.p_paddr, startPage));
-    wri(phdr.p_filesz, wri(phdr.p_memsz, shift));
+    wri(phdr.p_offset, phdrs.at(splitIndex).p_offset - splitShift - shift);
+    wri(phdr.p_paddr, phdrs.at(splitIndex).p_paddr - splitShift - shift);
+    wri(phdr.p_vaddr, phdrs.at(splitIndex).p_vaddr - splitShift - shift);
+    wri(phdr.p_filesz, wri(phdr.p_memsz, splitShift + shift));
     wri(phdr.p_flags, PF_R | PF_W);
     wri(phdr.p_align, getPageSize());
 }
@@ -638,6 +691,18 @@ void ElfFile<ElfFileParamNames>::writeReplacedSections(Elf_Off & curOff,
             }
         }
 
+        /* If there is .note.gnu.property section, then the PT_GNU_PROPERTY
+           segment must be sync'ed with it. */
+        if (sectionName == ".note.gnu.property") {
+            for (auto & phdr : phdrs) {
+                if (rdi(phdr.p_type) == PT_GNU_PROPERTY) {
+                    phdr.p_offset = shdr.sh_offset;
+                    phdr.p_vaddr = phdr.p_paddr = shdr.sh_addr;
+                    phdr.p_filesz = phdr.p_memsz = shdr.sh_size;
+                }
+            }
+        }
+
         curOff += roundUp(i.second.size(), sectionAlignment);
     }
 
@@ -675,7 +740,7 @@ void ElfFile<ElfFileParamNames>::rewriteSectionsLibrary()
     /* Some sections may already be replaced so account for that */
     unsigned int i = 1;
     Elf_Addr pht_size = sizeof(Elf_Ehdr) + (phdrs.size() + num_notes + 1)*sizeof(Elf_Phdr);
-    while( rdi(shdrs.at(i).sh_offset) <= pht_size && i < rdi(hdr()->e_shnum) ) {
+    while( i < rdi(hdr()->e_shnum) && rdi(shdrs.at(i).sh_offset) <= pht_size ) {
         if (not haveReplacedSection(getSectionName(shdrs.at(i))))
             replaceSection(getSectionName(shdrs.at(i)), rdi(shdrs.at(i).sh_size));
         i++;
@@ -829,7 +894,6 @@ void ElfFile<ElfFileParamNames>::rewriteSectionsExecutable()
        file by the minimum number of pages and adjust internal
        offsets. */
     if (neededSpace > startOffset) {
-
         /* We also need an additional program header, so adjust for that. */
         neededSpace += sizeof(Elf_Phdr);
         debug("needed space is %d\n", neededSpace);
@@ -839,10 +903,10 @@ void ElfFile<ElfFileParamNames>::rewriteSectionsExecutable()
         if (neededPages * getPageSize() > firstPage)
             error("virtual address space underrun!");
 
+        shiftFile(neededPages, startOffset);
+
         firstPage -= neededPages * getPageSize();
         startOffset += neededPages * getPageSize();
-
-        shiftFile(neededPages, firstPage);
     }
 
 
@@ -880,6 +944,14 @@ void ElfFile<ElfFileParamNames>::normalizeNoteSegments()
         size_t start_off = rdi(phdr.p_offset);
         size_t curr_off = start_off;
         size_t end_off = start_off + rdi(phdr.p_filesz);
+
+        /* Binaries produced by older patchelf versions may contain empty PT_NOTE segments.
+           For backwards compatibility, if we find one we should ignore it. */
+        bool empty = std::none_of(shdrs.begin(), shdrs.end(),
+            [&](auto & shdr) { return rdi(shdr.sh_offset) >= start_off && rdi(shdr.sh_offset) < end_off; });
+        if (empty)
+            continue;
+
         while (curr_off < end_off) {
             /* Find a section that starts at the current offset. If we can't
                find one, it means the SHT_NOTE sections weren't contiguous
@@ -1104,6 +1176,68 @@ std::string ElfFile<ElfFileParamNames>::getInterpreter()
     return std::string((char *) fileContents->data() + rdi(shdr.sh_offset), rdi(shdr.sh_size) - 1);
 }
 
+template<ElfFileParams>
+void ElfFile<ElfFileParamNames>::modifyOsAbi(osAbiMode op, const std::string & newOsAbi)
+{
+    unsigned char abi = hdr()->e_ident[EI_OSABI];
+
+    if (op == printOsAbi) {
+        switch (abi) {
+            case 0:  printf("System V\n"); break;
+            case 1:  printf("HP-UX\n"); break;
+            case 2:  printf("NetBSD\n"); break;
+            case 3:  printf("Linux\n"); break;
+            case 4:  printf("GNU Hurd\n"); break;
+            case 6:  printf("Solaris\n"); break;
+            case 7:  printf("AIX\n"); break;
+            case 8:  printf("IRIX\n"); break;
+            case 9:  printf("FreeBSD\n"); break;
+            case 10: printf("Tru64\n"); break;
+            case 12: printf("OpenBSD\n"); break;
+            case 13: printf("OpenVMS\n"); break;
+            default: printf("0x%02X\n", (unsigned int) abi);
+        }
+        return;
+    }
+
+    unsigned char newAbi;
+    std::string nabi = downcase(trim(newOsAbi));
+    if (nabi == "system v" || nabi == "system-v" || nabi == "sysv")
+        newAbi = 0;
+    else if (nabi == "hp-ux")
+        newAbi = 1;
+    else if (nabi == "netbsd")
+        newAbi = 2;
+    else if (nabi == "linux" || nabi == "gnu")
+        newAbi = 3;
+    else if (nabi == "gnu hurd" || nabi == "gnu-hurd" || nabi == "hurd")
+        newAbi = 4;
+    else if (nabi == "solaris")
+        newAbi = 6;
+    else if (nabi == "aix")
+        newAbi = 7;
+    else if (nabi == "irix")
+        newAbi = 8;
+    else if (nabi == "freebsd")
+        newAbi = 9;
+    else if (nabi == "tru64")
+        newAbi = 10;
+    else if (nabi == "openbsd")
+        newAbi = 12;
+    else if (nabi == "openvms")
+        newAbi = 13;
+    else
+        error("unrecognized OS ABI");
+
+    if (newAbi == abi) {
+        debug("current and requested OS ABIs are equal\n");
+        return;
+    }
+
+    hdr()->e_ident[EI_OSABI] = newAbi;
+    changed = true;
+}
+
 template<ElfFileParams>
 void ElfFile<ElfFileParamNames>::modifySoname(sonameMode op, const std::string & newSoname)
 {
@@ -1295,6 +1429,7 @@ void ElfFile<ElfFileParamNames>::modifyRPath(RPathOp op,
        string. */
     std::vector<std::string> neededLibs;
     auto dyn = (Elf_Dyn *)(fileContents->data() + rdi(shdrDynamic.sh_offset));
+    checkPointer(fileContents, dyn, sizeof(*dyn));
     Elf_Dyn *dynRPath = nullptr, *dynRunPath = nullptr;
     char * rpath = nullptr;
     for ( ; rdi(dyn->d_tag) != DT_NULL; dyn++) {
@@ -1739,6 +1874,9 @@ void ElfFile<ElfFileParamNames>::clearSymbolVersions(const std::set<std::string>
 }
 
 static bool printInterpreter = false;
+static bool printOsAbi = false;
+static bool setOsAbi = false;
+static std::string newOsAbi;
 static bool printSoname = false;
 static bool setSoname = false;
 static std::string newSoname;
@@ -1764,6 +1902,12 @@ static void patchElf2(ElfFile && elfFile, const FileContents & fileContents, con
     if (printInterpreter)
         printf("%s\n", elfFile.getInterpreter().c_str());
 
+    if (printOsAbi)
+        elfFile.modifyOsAbi(elfFile.printOsAbi, "");
+
+    if (setOsAbi)
+        elfFile.modifyOsAbi(elfFile.replaceOsAbi, newOsAbi);
+
     if (printSoname)
         elfFile.modifySoname(elfFile.printSoname, "");
 
@@ -1839,6 +1983,8 @@ void showHelp(const std::string & progName)
   [--set-interpreter FILENAME]\n\
   [--page-size SIZE]\n\
   [--print-interpreter]\n\
+  [--print-os-abi]\t\tPrints 'EI_OSABI' field of ELF header\n\
+  [--set-os-abi ABI]\t\tSets 'EI_OSABI' field of ELF header to ABI.\n\
   [--print-soname]\t\tPrints 'DT_SONAME' entry of .dynamic section. Raises an error if DT_SONAME doesn't exist\n\
   [--set-soname SONAME]\t\tSets 'DT_SONAME' entry to SONAME.\n\
   [--set-rpath RPATH]\n\
@@ -1888,6 +2034,14 @@ int mainWrapped(int argc, char * * argv)
         else if (arg == "--print-interpreter") {
             printInterpreter = true;
         }
+        else if (arg == "--print-os-abi") {
+            printOsAbi = true;
+        }
+        else if (arg == "--set-os-abi") {
+            if (++i == argc) error("missing argument");
+            setOsAbi = true;
+            newOsAbi = resolveArgument(argv[i]);
+        }
         else if (arg == "--print-soname") {
             printSoname = true;
         }
